@@ -71,7 +71,7 @@ static apr_status_t dosub(sed_eval_t *eval, char *rhsbuf, int n,
 static char *place(sed_eval_t *eval, char *asp, char *al1, char *al2);
 static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                             step_vars_storage *step_vars);
-static apr_status_t wline(sed_eval_t *eval, char *buf, int sz);
+static apr_status_t wline(sed_eval_t *eval, char *buf, apr_size_t sz);
 static apr_status_t arout(sed_eval_t *eval);
 
 static void eval_errf(sed_eval_t *eval, const char *fmt, ...)
@@ -92,11 +92,11 @@ static void eval_errf(sed_eval_t *eval, const char *fmt, ...)
  * grow_buffer
  */
 static void grow_buffer(apr_pool_t *pool, char **buffer,
-                        char **spend, unsigned int *cursize,
-                        unsigned int newsize)
+                        char **spend, apr_size_t *cursize,
+                        apr_size_t newsize)
 {
     char* newbuffer = NULL;
-    int spendsize = 0;
+    apr_size_t spendsize = 0;
     if (*cursize >= newsize)
         return;
     /* Avoid number of times realloc is called. It could cause huge memory
@@ -106,7 +106,7 @@ static void grow_buffer(apr_pool_t *pool, char **buffer,
     }
 
     /* Align it to 4 KB boundary */
-    newsize = (newsize  + ((1 << 12) - 1)) & ~((1 << 12) -1);
+    newsize = (newsize  + ((1 << 12) - 1)) & ~((1 << 12) - 1);
     newbuffer = apr_pcalloc(pool, newsize);
     if (*spend && *buffer && (*cursize > 0)) {
         spendsize = *spend - *buffer;
@@ -124,7 +124,7 @@ static void grow_buffer(apr_pool_t *pool, char **buffer,
 /*
  * grow_line_buffer
  */
-static void grow_line_buffer(sed_eval_t *eval, int newsize)
+static void grow_line_buffer(sed_eval_t *eval, apr_size_t newsize)
 {
     grow_buffer(eval->pool, &eval->linebuf, &eval->lspend,
                 &eval->lsize, newsize);
@@ -133,7 +133,7 @@ static void grow_line_buffer(sed_eval_t *eval, int newsize)
 /*
  * grow_hold_buffer
  */
-static void grow_hold_buffer(sed_eval_t *eval, int newsize)
+static void grow_hold_buffer(sed_eval_t *eval, apr_size_t newsize)
 {
     grow_buffer(eval->pool, &eval->holdbuf, &eval->hspend,
                 &eval->hsize, newsize);
@@ -142,7 +142,7 @@ static void grow_hold_buffer(sed_eval_t *eval, int newsize)
 /*
  * grow_gen_buffer
  */
-static void grow_gen_buffer(sed_eval_t *eval, int newsize,
+static void grow_gen_buffer(sed_eval_t *eval, apr_size_t newsize,
                             char **gspend)
 {
     if (gspend == NULL) {
@@ -156,9 +156,9 @@ static void grow_gen_buffer(sed_eval_t *eval, int newsize,
 /*
  * appendmem_to_linebuf
  */
-static void appendmem_to_linebuf(sed_eval_t *eval, const char* sz, int len)
+static void appendmem_to_linebuf(sed_eval_t *eval, const char* sz, apr_size_t len)
 {
-    unsigned int reqsize = (eval->lspend - eval->linebuf) + len;
+    apr_size_t reqsize = (eval->lspend - eval->linebuf) + len;
     if (eval->lsize < reqsize) {
         grow_line_buffer(eval, reqsize);
     }
@@ -169,21 +169,36 @@ static void appendmem_to_linebuf(sed_eval_t *eval, const char* sz, int len)
 /*
  * append_to_linebuf
  */
-static void append_to_linebuf(sed_eval_t *eval, const char* sz)
+static void append_to_linebuf(sed_eval_t *eval, const char* sz,
+                              step_vars_storage *step_vars)
 {
-    int len = strlen(sz);
+    apr_size_t len = strlen(sz);
+    char *old_linebuf = eval->linebuf;
     /* Copy string including null character */
     appendmem_to_linebuf(eval, sz, len + 1);
     --eval->lspend; /* lspend will now point to NULL character */
+    /* Sync step_vars after a possible linebuf expansion */
+    if (step_vars && old_linebuf != eval->linebuf) {
+        if (step_vars->loc1) {
+            step_vars->loc1 = step_vars->loc1 - old_linebuf + eval->linebuf;
+        }
+        if (step_vars->loc2) {
+            step_vars->loc2 = step_vars->loc2 - old_linebuf + eval->linebuf;
+        }
+        if (step_vars->locs) {
+            step_vars->locs = step_vars->locs - old_linebuf + eval->linebuf;
+        }
+    }
 }
 
 /*
  * copy_to_linebuf
  */
-static void copy_to_linebuf(sed_eval_t *eval, const char* sz)
+static void copy_to_linebuf(sed_eval_t *eval, const char* sz,
+                            step_vars_storage *step_vars)
 {
     eval->lspend = eval->linebuf;
-    append_to_linebuf(eval, sz);
+    append_to_linebuf(eval, sz, step_vars);
 }
 
 /*
@@ -191,8 +206,8 @@ static void copy_to_linebuf(sed_eval_t *eval, const char* sz)
  */
 static void append_to_holdbuf(sed_eval_t *eval, const char* sz)
 {
-    int len = strlen(sz);
-    unsigned int reqsize = (eval->hspend - eval->holdbuf) + len + 1;
+    apr_size_t len = strlen(sz);
+    apr_size_t reqsize = (eval->hspend - eval->holdbuf) + len + 1;
     if (eval->hsize <= reqsize) {
         grow_hold_buffer(eval, reqsize);
     }
@@ -215,8 +230,8 @@ static void copy_to_holdbuf(sed_eval_t *eval, const char* sz)
  */
 static void append_to_genbuf(sed_eval_t *eval, const char* sz, char **gspend)
 {
-    int len = strlen(sz);
-    unsigned int reqsize = (*gspend - eval->genbuf) + len + 1;
+    apr_size_t len = strlen(sz);
+    apr_size_t reqsize = (*gspend - eval->genbuf) + len + 1;
     if (eval->gsize < reqsize) {
         grow_gen_buffer(eval, reqsize, gspend);
     }
@@ -230,8 +245,8 @@ static void append_to_genbuf(sed_eval_t *eval, const char* sz, char **gspend)
  */
 static void copy_to_genbuf(sed_eval_t *eval, const char* sz)
 {
-    int len = strlen(sz);
-    unsigned int reqsize = len + 1;
+    apr_size_t len = strlen(sz);
+    apr_size_t reqsize = len + 1;
     if (eval->gsize < reqsize) {
         grow_gen_buffer(eval, reqsize, NULL);
     }
@@ -353,7 +368,7 @@ apr_status_t sed_eval_file(sed_eval_t *eval, apr_file_t *fin, void *fout)
 /*
  * sed_eval_buffer
  */
-apr_status_t sed_eval_buffer(sed_eval_t *eval, const char *buf, int bufsz, void *fout)
+apr_status_t sed_eval_buffer(sed_eval_t *eval, const char *buf, apr_size_t bufsz, void *fout)
 {
     apr_status_t rv;
 
@@ -383,7 +398,7 @@ apr_status_t sed_eval_buffer(sed_eval_t *eval, const char *buf, int bufsz, void
 
     while (bufsz) {
         char *n;
-        int llen;
+        apr_size_t llen;
 
         n = memchr(buf, '\n', bufsz);
         if (n == NULL)
@@ -442,7 +457,7 @@ apr_status_t sed_finalize_eval(sed_eval_t *eval, void *fout)
              * buffer is not a newline.
              */
             /* Assure space for NULL */
-            append_to_linebuf(eval, "");
+            append_to_linebuf(eval, "", NULL);
         }
 
         *eval->lspend = '\0';
@@ -586,8 +601,8 @@ static int match(sed_eval_t *eval, char *expbuf, int gf,
     char   *p1;
     int circf;
 
-    if(gf) {
-        if(*expbuf)    return(0);
+    if (gf) {
+        if (*expbuf)    return(0);
         step_vars->locs = p1 = step_vars->loc2;
     } else {
         p1 = eval->linebuf;
@@ -604,16 +619,16 @@ static int match(sed_eval_t *eval, char *expbuf, int gf,
 static int substitute(sed_eval_t *eval, sed_reptr_t *ipc,
                       step_vars_storage *step_vars)
 {
-    if(match(eval, ipc->re1, 0, step_vars) == 0)    return(0);
+    if (match(eval, ipc->re1, 0, step_vars) == 0)    return(0);
 
     eval->numpass = 0;
     eval->sflag = 0;        /* Flags if any substitution was made */
     if (dosub(eval, ipc->rhs, ipc->gfl, step_vars) != APR_SUCCESS)
         return -1;
 
-    if(ipc->gfl) {
-        while(*step_vars->loc2) {
-            if(match(eval, ipc->re1, 1, step_vars) == 0) break;
+    if (ipc->gfl) {
+        while (*step_vars->loc2) {
+            if (match(eval, ipc->re1, 1, step_vars) == 0) break;
             if (dosub(eval, ipc->rhs, ipc->gfl, step_vars) != APR_SUCCESS)
                 return -1;
         }
@@ -631,9 +646,9 @@ static apr_status_t dosub(sed_eval_t *eval, char *rhsbuf, int n,
     int c;
     apr_status_t rv = APR_SUCCESS;
 
-    if(n > 0 && n < 999) {
+    if (n > 0 && n < 999) {
         eval->numpass++;
-        if(n != eval->numpass) return APR_SUCCESS;
+        if (n != eval->numpass) return APR_SUCCESS;
     }
     eval->sflag = 1;
     lp = eval->linebuf;
@@ -666,7 +681,7 @@ static apr_status_t dosub(sed_eval_t *eval, char *rhsbuf, int n,
     lp = step_vars->loc2;
     step_vars->loc2 = sp - eval->genbuf + eval->linebuf;
     append_to_genbuf(eval, lp, &sp);
-    copy_to_linebuf(eval, eval->genbuf);
+    copy_to_linebuf(eval, eval->genbuf, step_vars);
     return rv;
 }
 
@@ -676,8 +691,8 @@ static apr_status_t dosub(sed_eval_t *eval, char *rhsbuf, int n,
 static char *place(sed_eval_t *eval, char *asp, char *al1, char *al2)
 {
     char *sp = asp;
-    int n = al2 - al1;
-    unsigned int reqsize = (sp - eval->genbuf) + n + 1;
+    apr_size_t n = al2 - al1;
+    apr_size_t reqsize = (sp - eval->genbuf) + n + 1;
 
     if (eval->gsize < reqsize) {
         grow_gen_buffer(eval, reqsize, &sp);
@@ -703,7 +718,7 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
     switch(ipc->command) {
 
         case ACOM:
-            if(eval->aptr >= &eval->abuf[SED_ABUFSIZE]) {
+            if (eval->aptr >= &eval->abuf[SED_ABUFSIZE]) {
                 eval_errf(eval, SEDERR_TMAMES, eval->lnum);
             } else {
                 *eval->aptr++ = ipc;
@@ -713,27 +728,29 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
 
         case CCOM:
             eval->delflag = 1;
-            if(!eval->inar[ipc->nrep] || eval->dolflag) {
+            if (!eval->inar[ipc->nrep] || eval->dolflag) {
                 for (p1 = ipc->re1; *p1; p1++)
                     ;
                 rv = wline(eval, ipc->re1, p1 - ipc->re1);
             }
             break;
+
         case DCOM:
             eval->delflag++;
             break;
+
         case CDCOM:
             p1 = eval->linebuf;
 
-            while(*p1 != '\n') {
-                if(*p1++ == 0) {
+            while (*p1 != '\n') {
+                if (*p1++ == 0) {
                     eval->delflag++;
                     return APR_SUCCESS;
                 }
             }
 
             p1++;
-            copy_to_linebuf(eval, p1);
+            copy_to_linebuf(eval, p1, step_vars);
             eval->jflag++;
             break;
 
@@ -743,12 +760,12 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
             break;
 
         case GCOM:
-            copy_to_linebuf(eval, eval->holdbuf);
+            copy_to_linebuf(eval, eval->holdbuf, step_vars);
             break;
 
         case CGCOM:
-            append_to_linebuf(eval, "\n");
-            append_to_linebuf(eval, eval->holdbuf);
+            append_to_linebuf(eval, "\n", step_vars);
+            append_to_linebuf(eval, eval->holdbuf, step_vars);
             break;
 
         case HCOM:
@@ -769,17 +786,16 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
             eval->jflag = 1;
             break;
 
-
         case LCOM:
             p1 = eval->linebuf;
             p2 = eval->genbuf;
             eval->genbuf[72] = 0;
-            while(*p1)
-                if((unsigned char)*p1 >= 040) {
-                    if(*p1 == 0177) {
+            while (*p1) {
+                if ((unsigned char)*p1 >= 040) {
+                    if (*p1 == 0177) {
                         p3 = rub;
                         while ((*p2++ = *p3++) != 0)
-                            if(p2 >= eval->lcomend) {
+                            if (p2 >= eval->lcomend) {
                                 *p2 = '\\';
                                 rv = wline(eval, eval->genbuf,
                                            strlen(eval->genbuf));
@@ -791,9 +807,9 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                         p1++;
                         continue;
                     }
-                    if(!isprint(*p1 & 0377)) {
+                    if (!isprint(*p1 & 0377)) {
                         *p2++ = '\\';
-                        if(p2 >= eval->lcomend) {
+                        if (p2 >= eval->lcomend) {
                             *p2 = '\\';
                             rv = wline(eval, eval->genbuf,
                                        strlen(eval->genbuf));
@@ -802,7 +818,7 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                             p2 = eval->genbuf;
                         }
                         *p2++ = (*p1 >> 6) + '0';
-                        if(p2 >= eval->lcomend) {
+                        if (p2 >= eval->lcomend) {
                             *p2 = '\\';
                             rv = wline(eval, eval->genbuf,
                                        strlen(eval->genbuf));
@@ -811,7 +827,7 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                             p2 = eval->genbuf;
                         }
                         *p2++ = ((*p1 >> 3) & 07) + '0';
-                        if(p2 >= eval->lcomend) {
+                        if (p2 >= eval->lcomend) {
                             *p2 = '\\';
                             rv = wline(eval, eval->genbuf,
                                        strlen(eval->genbuf));
@@ -820,7 +836,7 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                             p2 = eval->genbuf;
                         }
                         *p2++ = (*p1++ & 07) + '0';
-                        if(p2 >= eval->lcomend) {
+                        if (p2 >= eval->lcomend) {
                             *p2 = '\\';
                             rv = wline(eval, eval->genbuf,
                                        strlen(eval->genbuf));
@@ -830,7 +846,7 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                         }
                     } else {
                         *p2++ = *p1++;
-                        if(p2 >= eval->lcomend) {
+                        if (p2 >= eval->lcomend) {
                             *p2 = '\\';
                             rv = wline(eval, eval->genbuf,
                                        strlen(eval->genbuf));
@@ -842,7 +858,7 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                 } else {
                     p3 = trans[(unsigned char)*p1-1];
                     while ((*p2++ = *p3++) != 0)
-                        if(p2 >= eval->lcomend) {
+                        if (p2 >= eval->lcomend) {
                             *p2 = '\\';
                             rv = wline(eval, eval->genbuf,
                                        strlen(eval->genbuf));
@@ -853,39 +869,41 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                     p2--;
                     p1++;
                 }
+            }
             *p2 = 0;
             rv = wline(eval, eval->genbuf, strlen(eval->genbuf));
             break;
 
         case NCOM:
-            if(!eval->commands->nflag) {
+            if (!eval->commands->nflag) {
                 rv = wline(eval, eval->linebuf, eval->lspend - eval->linebuf);
                 if (rv != APR_SUCCESS)
                     return rv;
             }
 
-            if(eval->aptr > eval->abuf) {
+            if (eval->aptr > eval->abuf) {
                 rv = arout(eval);
                 if (rv != APR_SUCCESS)
                     return rv;
             }
             eval->lspend = eval->linebuf;
             eval->pending = ipc->next;
-
             break;
+
         case CNCOM:
-            if(eval->aptr > eval->abuf) {
+            if (eval->aptr > eval->abuf) {
                 rv = arout(eval);
                 if (rv != APR_SUCCESS)
                     return rv;
             }
-            append_to_linebuf(eval, "\n");
+            append_to_linebuf(eval, "\n", step_vars);
             eval->pending = ipc->next;
             break;
 
         case PCOM:
             rv = wline(eval, eval->linebuf, eval->lspend - eval->linebuf);
             break;
+
         case CPCOM:
             for (p1 = eval->linebuf; *p1 != '\n' && *p1 != '\0'; p1++);
             rv = wline(eval, eval->linebuf, p1 - eval->linebuf);
@@ -898,7 +916,7 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                     break;
             }
 
-            if(eval->aptr > eval->abuf) {
+            if (eval->aptr > eval->abuf) {
                 rv = arout(eval);
                 if (rv != APR_SUCCESS)
                     return rv;
@@ -906,8 +924,9 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
 
             eval->quitflag = 1;
             break;
+
         case RCOM:
-            if(eval->aptr >= &eval->abuf[SED_ABUFSIZE]) {
+            if (eval->aptr >= &eval->abuf[SED_ABUFSIZE]) {
                 eval_errf(eval, SEDERR_TMRMES, eval->lnum);
             } else {
                 *eval->aptr++ = ipc;
@@ -920,8 +939,8 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
             if (i == -1) {
                 return APR_EGENERAL;
             }
-            if(ipc->pfl && eval->commands->nflag && i) {
-                if(ipc->pfl == 1) {
+            if (ipc->pfl && eval->commands->nflag && i) {
+                if (ipc->pfl == 1) {
                     rv = wline(eval, eval->linebuf, eval->lspend -
                                eval->linebuf);
                     if (rv != APR_SUCCESS)
@@ -939,7 +958,7 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
             break;
 
         case TCOM:
-            if(eval->sflag == 0)  break;
+            if (eval->sflag == 0)  break;
             eval->sflag = 0;
             eval->jflag = 1;
             break;
@@ -949,16 +968,17 @@ static apr_status_t command(sed_eval_t *eval, sed_reptr_t *ipc,
                 apr_file_printf(eval->fcode[ipc->findex], "%s\n",
                                 eval->linebuf);
             break;
+
         case XCOM:
             copy_to_genbuf(eval, eval->linebuf);
-            copy_to_linebuf(eval, eval->holdbuf);
+            copy_to_linebuf(eval, eval->holdbuf, step_vars);
             copy_to_holdbuf(eval, eval->genbuf);
             break;
 
         case YCOM:
             p1 = eval->linebuf;
             p2 = ipc->re1;
-            while((*p1 = p2[(unsigned char)*p1]) != 0)    p1++;
+            while ((*p1 = p2[(unsigned char)*p1]) != 0)    p1++;
             break;
     }
     return rv;
@@ -1008,7 +1028,7 @@ static apr_status_t arout(sed_eval_t *eval)
 /*
  * wline
  */
-static apr_status_t wline(sed_eval_t *eval, char *buf, int sz)
+static apr_status_t wline(sed_eval_t *eval, char *buf, apr_size_t sz)
 {
     apr_status_t rv = APR_SUCCESS;
     rv = eval->writefn(eval->fout, buf, sz);
